<?php
/*
 * Plugin Name: Truevo Payments
 * Plugin URI: https://truevo.com/
 * Description: Take online payments with the Truevo Payment Plugin. Tokenisation lets your customers save their card details for repeat purchase convenience. Perform reversals, refunds and captures.  PCI compliant and 3DS v2 enabled, to keep you and your customers safe.
 * Version: 1.7.1
 * Author: Truevo Payments
 * Author URI: https://truevo.com/
 * Text Domain: truevo-gateway-woocommerce
 * Domain Path: /languages
*
 
 /*
 * This action hook registers our PHP class as a WooCommerce payment gateway
 */


// Make sure WooCommerce is active
if (!in_array('woocommerce/woocommerce.php', apply_filters('active_plugins', get_option('active_plugins')))) {
    return;
}

include('includes/truevo-gateway-api.php');

add_filter('woocommerce_payment_gateways', 'truevo_add_gateway_class');

/**
 * truevo_add_gateway_class function.
 *
 * @access public
 *
 * @param mixed $gateways
 *
 * @return void
 */
function truevo_add_gateway_class($gateways)
{
    $gateways[] = 'WC_Truevo_Gateway';

    return $gateways;
}

if (!function_exists('truevo_gateway_action_links')) {

    /**
     * truevo_gateway_action_links function.
     *
     * @access public
     *
     * @param mixed $links
     *
     * @return void
     */
    function truevo_gateway_action_links($links)
    {
        $plugin_links = array();

        if (function_exists('WC')) {
            if (version_compare(WC()->version, '2.6', '>=')) {
                $section_slug = 'truevo_gateway';
            } else {
                $section_slug = strtolower('WC_Truevo_Gateway');
            }
            $setting_url = admin_url('admin.php?page=wc-settings&tab=checkout&section=' . $section_slug);
            $plugin_links[] = '<a href="' . esc_url($setting_url) . '">' . esc_html__('Settings (widget)', 'truevo-gateway-woocommerce') . '</a>';
        }

        return array_merge($plugin_links, $links);
    }

    add_filter('plugin_action_links_' . plugin_basename(__FILE__), 'truevo_gateway_action_links');
}

/**
 * Add a custom action to order actions select box on edit order page
 * Only added for capture payments
 *
 * @param array $actions order actions array to display
 * @return array - updated actions
 */
function truevo_wc_add_order_meta_box_action($actions)
{
    global $theorder;

    if ($theorder->get_payment_method() !== 'truevo_gateway') {
        return;
    }

    $truevo_payment_captured = get_post_meta($theorder->id, 'truevo_payment_captured', true);
    $truevo_payment_id = get_post_meta($theorder->id, 'truevo_payment_transactionId', true);
    $truevo_payment_type = get_post_meta($theorder->id, 'truevo_payment_transaction_type', true);

    if ($truevo_payment_captured != 'yes' && $truevo_payment_type == 'authorize') {
        $truevoapi = truevo_api_connection();
        $get_payment_details = $truevoapi->get_payment_details($truevo_payment_id, true);

        if ($get_payment_details === 'capture') {
            $actions['wc_truevo_order_action'] = __('Capture the payment', 'truevo-gateway-woocommerce');
        }
    }

    return $actions;
}


/**
 * Add an order note when capture action is clicked
 *
 * @param \WC_Order $order
 */
function truevo_wc_process_order_meta_box_action($order)
{

    $order_id = $order->id;

    $truevo_payment_id = get_post_meta($order_id, 'truevo_payment_transactionId', true);
    $truevo_payment_rrn = get_post_meta($order_id, 'truevo_payment_rrn', true);
    $truevo_payment_captured = get_post_meta($order_id, 'truevo_payment_captured', true);
    $truevo_payment_type = get_post_meta($order_id, 'truevo_payment_transaction_type', true);

    $log = new WC_Logger();
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    if (!empty($truevo_payment_id) && $truevo_payment_rrn && $truevo_payment_captured != 'yes' && $truevo_payment_type == 'authorize') {

        $order_reference = $order->get_id();

        $data_info = array(
            'transactionUniqueIdentifier' => $truevo_payment_id,
            'retrievalReferenceNumber' => $truevo_payment_rrn,
            'amount' => number_format($order->get_total(), 2, '.', ''),
            'merchantReference' => "$order_reference",
            'paymentReference' => $truevo_gateway_settings['store_name'],
            'billing_address' => $order->get_billing_address_1(),
            'billing_address_2' => $order->get_billing_address_2(),
            'billing_city' => $order->get_billing_city(),
            'billing_state' => $order->get_billing_state(),
            'billing_country' => $order->get_billing_country(),
            'billing_postcode' => $order->get_billing_postcode(),
            'billing_first_name' => $order->billing_first_name,
            'billing_last_name' => $order->billing_last_name,
            'delivery_address' => $order->get_shipping_address_1(),
            'delivery_address_2' => $order->get_shipping_address_2(),
            'delivery_city' => $order->get_shipping_city(),
            'delivery_country' => $order->get_shipping_country(),
            'delivery_postcode' => $order->get_shipping_postcode(),
            'delivery_state' => $order->get_shipping_state(),
            'email' => $order->get_billing_email(),
            'user_ip' => truevo_get_client_ip(),
        );

        if ($truevo_gateway_settings['debug_log'] == 'yes') {
            $log->add('truevo_gateway', 'Truevo payments capture request. ' . print_r($data_info, true));
        }
        $truevoapi = truevo_api_connection();
        $process_capture = $truevoapi->process_truevo_capture($data_info);

        if ($truevo_gateway_settings['debug_log'] == 'yes') {
            $log->add('truevo_gateway', 'Truevo payments capture response. ' . print_r($process_capture, true));
        }
        $rejected = true;

        if ($process_capture['code'] == '200' && !empty($process_capture['result'])) {

            if ($process_capture['result']->outcome->status === 'RECEIVED') {

                $transaction_id = $process_capture['result']->transactionUniqueIdentifier;

                $query_final_response = $truevoapi->poll_transaction_status($transaction_id, $truevo_gateway_settings['debug_log']);

                $query_final_status = $query_final_response['status'] ?? '';
                $query_final_message = $query_final_response['reason'] ?? '';

                if ($query_final_status === 'COMPLETE') {
                    
                    $transaction_id = $process_capture['result']->transactionUniqueIdentifier;
                    $order->update_status('processing');

                    $order->add_order_note('Truevo Payments: Your payment was successfully captured.', false);

                    update_post_meta($order_id, 'truevo_payment_captured', 'yes');
                    update_post_meta($order_id, 'truevo_payment_captured_status', $query_final_status);
                    update_post_meta($order_id, 'truevo_payment_transactionId', $transaction_id);

                    $order->save();

                    $rejected = false;
                } else {

                    update_post_meta($order_id, 'truevo_payment_captured_status', $query_final_status);
                    $order->add_order_note($query_final_message, false);
                }


            }
        }

        if ($rejected) {

            $order->update_status('on-hold');
            $order->add_order_note('Truevo Payments: Payment capture could not be processed.', false);
        }

        $order->save();
    }
}

add_action('woocommerce_order_actions', 'truevo_wc_add_order_meta_box_action');
add_action('woocommerce_order_action_wc_truevo_order_action', 'truevo_wc_process_order_meta_box_action');


add_action('admin_head', 'truevo_hide_show_wc_refund_reverse_buttons');
/**
 * truevo_hide_show_wc_refund_reverse_buttons function
 * @return void
 */
function truevo_hide_show_wc_refund_reverse_buttons()
{

    global $post;
    if (strpos($_SERVER['REQUEST_URI'], 'post.php?post=') === false) {
        return;
    }

    if (empty($post) || $post->post_type != 'shop_order') {
        return;
    }

    $order_id = $post->ID;

    $order = wc_get_order($order_id);

    if ($order->get_payment_method() !== 'truevo_gateway') {
        return;
    }

    $truevo_payment_id = get_post_meta($order_id, 'truevo_payment_transactionId', true);
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

    if ($truevo_payment_id) {
        
        $truevoapi = truevo_api_connection();
        $refund_mode = $truevoapi->get_payment_details( $truevo_payment_id, false, $truevo_gateway_settings['debug_log'] );
        if ($refund_mode == "reverse") {
            ?>
            <style type="text/css">
                .refund-items {
                    display: none !important
                }
            </style>
            <?php
        } else if ($refund_mode == "refund") {
            ?>
            <style type="text/css">
                .reverse-items {
                    display: none !important
                }
            </style>
            <?php
        } else {
            ?>
            <style type="text/css">
                .refund-items, .reverse-items {
                    display: none !important
                }
            </style>
            <?php
        }
    }

}

add_action('plugins_loaded', 'truevo_init_gateway_class');
/**
 * truevo_init_gateway_class function.
 *
 * @access public
 * @return void
 */
function truevo_init_gateway_class()
{

    class WC_Truevo_Gateway extends WC_Payment_Gateway
    {

        /**
         * Class constructor, more about it in Step 3
         */
        public function __construct()
        {

            $this->id = 'truevo_gateway';
            $this->has_fields = true;
            $this->method_title = 'Truevo Payments';
            $this->method_description = 'Accept payments from your customers with Truevo payments'; // will be displayed on the options page

            $this->supports = array(
                'products',
                'refunds'
            );

            // Method with all the options fields
            $this->init_form_fields();

            // Load the settings.
            $this->init_settings();
            $this->title = $this->get_option('title');
            $this->description = $this->get_option('description');
            $this->enabled = $this->get_option('enabled');
            $this->store_name = $this->get_option('store_name');
            $this->show_logo = $this->get_option('show_logo');
            $this->widget_theme = $this->get_option('widget_theme');
            $this->debug_log = $this->get_option('debug_log');

            $this->testmode = ('yes' === $this->get_option('testmode')) ? 'test' : 'live';
            $this->merchant_id = ($this->testmode == 'test') ? $this->get_option('test_merchant_id') : $this->get_option('live_merchant_id');
            $this->terminal_id = ($this->testmode == 'test') ? $this->get_option('test_terminal_id') : $this->get_option('live_terminal_id');

            $this->api_url = ($this->testmode == 'test') ? 'https://sandbox.unify.truevo.com' : 'https://unify.truevo.com';

            $this->enable_authorization = $this->get_option('enable_authorization');
            $this->token_authorization = $this->get_option('token_authorization');
            $this->enable_tokenization = $this->get_option('enable_tokenization');


            $this->css_styles = $this->get_option('css_styles');

            $this->truevoapi = new Truevo_API_Wrapper($this->api_url, $this->token_authorization, $this->merchant_id, $this->terminal_id);

            $this->log = new WC_Logger();

            // This action hook saves the settings
            add_action('woocommerce_update_options_payment_gateways_' . $this->id, array(
                $this,
                'process_admin_options'
            ));

            add_action('woocommerce_order_item_add_action_buttons', array($this, 'reverse_button'));

            add_action('woocommerce_receipt_' . $this->id, array($this, 'receipt_page'));

            add_action('woocommerce_api_' . $this->id, array($this, 'truevo_gateway_webhook'));

            // We need custom JavaScript to obtain a token
            add_action('wp_enqueue_scripts', array($this, 'payment_scripts'));

        }


        public function reverse_button($order)
        {
            ?>
            <button type="button" class="button reverse-items"
                    data-order_id="<?php echo $order->get_id(); ?>"><?php esc_html_e('Reverse', 'woocommerce'); ?></button>
            <?php
        }


        /**
         * init_form_fields function.
         *
         * @access public
         * @return void
         */
        public function init_form_fields()
        {

            $this->form_fields = array(
                'enabled' => array(
                    'title' => __('Enable/Disable', 'truevo-gateway-woocommerce'),
                    'label' => __('Enable Truevo payments', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'description' => '',
                    'default' => 'no'
                ),
                'title' => array(
                    'title' => __('Title', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('This controls the title which the user sees during checkout.', 'truevo-gateway-woocommerce'),
                    'default' => __('Pay with card', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'description' => array(
                    'title' => __('Description', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('This controls the description which the user sees during checkout.', 'truevo-gateway-woocommerce'),
                    'default' => __('Pay with your card securely via Truevo payments.', 'truevo-gateway-woocommerce'),
                ),
                'store_name' => array(
                    'title' => __('Store Name', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('This is a store name which will show in widget', 'truevo-gateway-woocommerce'),
                    'default' => '',
                    'desc_tip' => true,
                ),
                'show_logo' => array(
                    'title' => __('Show Truevo Logo', 'truevo-gateway-woocommerce'),
                    'label' => __('Enable to show Truevo logo', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'description' => __('Show Truevo logo in the widget', 'truevo-gateway-woocommerce'),
                    'default' => 'yes',
                    'desc_tip' => true,
                ),
                'widget_theme' => array(
                    'title' => __('Widget Theme', 'truevo-gateway-woocommerce'),
                    'type' => 'select',
                    'options' => array(
                        'plan' => __('Plain', 'truevo-gateway-woocommerce'),
                        'card' => __('Card', 'truevo-gateway-woocommerce')
                    ),
                    'default' => 'plan',
                ),
                'testmode' => array(
                    'title' => __('Test mode', 'truevo-gateway-woocommerce'),
                    'label' => __('Enable Test Mode', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'description' => __('Place the payment gateway in test mode using test API keys.', 'truevo-gateway-woocommerce'),
                    'default' => 'yes',
                    'desc_tip' => true,
                ),
                'test_merchant_id' => array(
                    'title' => __('Test Merchant ID (mid)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('Sandbox Merchant ID (mid)', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'test_terminal_id' => array(
                    'title' => __('Test Terminal ID (tid)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('Sandbox Terminal ID (tid)', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'live_merchant_id' => array(
                    'title' => __('Live Merchant ID (mid)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('Live Merchant ID (mid)', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'live_terminal_id' => array(
                    'title' => __('Live Terminal ID (tid)', 'truevo-gateway-woocommerce'),
                    'type' => 'text',
                    'description' => __('Live Terminal ID (tid)', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'enable_authorization' => array(
                    'title' => __('Enable Pre-Authorisation', 'truevo-gateway-woocommerce'),
                    'label' => __('Enables the pre-authorisation and capture payment mode', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'description' => __('Enables the pre-authorisation and capture payment mode.', 'truevo-gateway-woocommerce'),
                    'default' => 'no',
                    'desc_tip' => true,
                ),
                'enable_tokenization' => array(
                    'title' => __('Enable Tokenization', 'truevo-gateway-woocommerce'),
                    'label' => __('Let’s your customers safely save their card details for repeat purchases.', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'description' => __('Let’s your customers safely save their card details for repeat purchases.', 'truevo-gateway-woocommerce'),
                    'default' => 'no',
                    'desc_tip' => true,
                ),
                'token_authorization' => array(
                    'title' => __('Token', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('Token for pre-authorizations', 'truevo-gateway-woocommerce'),
                    'desc_tip' => true,
                ),
                'css_styles' => array(
                    'title' => __('CSS Styles', 'truevo-gateway-woocommerce'),
                    'type' => 'textarea',
                    'description' => __('Custom CSS for widget.', 'truevo-gateway-woocommerce'),
                    'default' => '',
                ),
                'debug_log' => array(
                    'title' => __('Debug log', 'truevo-gateway-woocommerce'),
                    'type' => 'checkbox',
                    'label' => __('Enable Logging.', 'truevo-gateway-woocommerce'),
                    'default' => 'no',
                    'description' => sprintf(__('Log gateway events, stored in %s', 'truevo-gateway-woocommerce'), '<code>' . WC_Log_Handler_File::get_log_file_path('truevo_gateway') . '</code>')
                )
            );
        }


        /**
         * receipt_page function.
         *
         * @access public
         * @return void
         */
        public function receipt_page($order)
        {

            $order = wc_get_order($order);

            $log = new WC_Logger();

            $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

            if (!empty($truevo_gateway_settings)) {

                if (defined('ICL_LANGUAGE_CODE')) {
                    $language = ICL_LANGUAGE_CODE;
                } else {
                    $language = get_locale();
                }

                $language = explode("_", $language)[0];

                $enable_authorization = ('yes' === $truevo_gateway_settings['enable_authorization']) ? 'authorize' : 'sale';

                $testmode = ('yes' === $truevo_gateway_settings['testmode']) ? 'test' : 'live';
                $merchant_id = ($testmode == 'test') ? $truevo_gateway_settings['test_merchant_id'] : $truevo_gateway_settings['live_merchant_id'];
                $terminal_id = ($testmode == 'test') ? $truevo_gateway_settings['test_terminal_id'] : $truevo_gateway_settings['live_terminal_id'];

                $order_reference = $order->get_id();

                $order_user = $order->get_user();
                $order_username = '';
                if (!empty($order_user)) {
                    $order_username = $order_user->__get('user_login');
                } else {
                    $order_username = get_post_meta( $order->get_id(), '_billing_email', true );
                }
                
                $args = array(
                    'currency' => get_woocommerce_currency(),
                    'merchant_id' => $merchant_id,
                    'terminal_id' => $terminal_id,
                    'store_name' => $truevo_gateway_settings['store_name'],
                    'show_logo' => ($truevo_gateway_settings['show_logo'] == 'yes') ? true : false,
                    'widget_theme' => $truevo_gateway_settings['widget_theme'],
                    'css_styles' => $truevo_gateway_settings['css_styles'],
                    'amount' => number_format($order->get_total(), 2, '.', ''),
                    'billing_address' => $order->get_billing_address_1(),
                    'billing_address_2' => $order->get_billing_address_2(),
                    'billing_city' => $order->get_billing_city(),
                    'billing_state' => $order->get_billing_state(),
                    'billing_country' => $order->get_billing_country(),
                    'billing_postcode' => $order->get_billing_postcode(),
                    'billing_first_name' => $order->billing_first_name,
                    'billing_last_name' => $order->billing_last_name,
                    'callback_url' => WC()->api_request_url('WC_Truevo_Gateway'),
                    'customer_url' => WC()->api_request_url('WC_Truevo_Gateway'),
                    'delivery_address' => $order->get_shipping_address_1(),
                    'delivery_address_2' => $order->get_shipping_address_2(),
                    'delivery_city' => $order->get_shipping_city(),
                    'delivery_country' => $order->get_shipping_country(),
                    'delivery_postcode' => $order->get_shipping_postcode(),
                    'delivery_state' => $order->get_shipping_state(),
                    'email' => $order->get_billing_email(),
                    'order_reference' => "$order_reference",
                    'user_ip' => truevo_get_client_ip(),
                    'language' => strtoupper($language),
                    'source' => get_bloginfo('url'),
                    'transaction_type' => $enable_authorization
                );

                $configuration = [
                    'transactionType' => $args['transaction_type'],
                    'widgetConfiguration' => [
                        'language' => $args['language'],
                        'showTruevo' => $args['show_logo'],
                        'iframeStyles' => [
                            'theme' => $args['widget_theme'],
                            'style' => $args['css_styles'],
                        ]
                    ],
                    'transaction' => [
                        'transactionUniqueIdentifier' => '',
                        'merchantReference' => $args['order_reference'],
                        'paymentReference' => $args['store_name'],
                        'amount' => [
                            'totalAmount' => $args['amount'],
                            'currencyAlphaCode' => $args['currency'],
                            'details' => [
                                'paymentAmount' => $args['amount'],
                                'gratuityAmount' => 0,
                            ],
                        ],
                        'source' => 'ECOMMERCE'
                    ],
                    'merchant' => [
                        'descriptor' => $args['store_name'],
                        'mid' => $args['merchant_id'],
                        'tid' => $args['terminal_id']
                    ],
                    'additionalData' => [],
                    'customer' => [
                        'firstName' => $args['billing_first_name'],
                        'lastName' => $args['billing_last_name'],
                        'contact' => [
                            'email' => $args['email'],
                            'ip' => $args['user_ip'],
                        ],
                        'physicalAddress' => [
                            'street1' => $args['delivery_address'],
                            'street2' => $args['delivery_address_2'],
                            'city' => $args['delivery_city'],
                            'state' => $args['delivery_state'],
                            'postCodeZip' => $args['delivery_postcode'],
                            'countryAlphaCode' => $args['delivery_country'],
                        ],
                        'billingAddress' => [
                            'street1' => $args['billing_address'],
                            'street2' => $args['billing_address_2'],
                            'city' => $args['billing_city'],
                            'state' => $args['billing_state'],
                            'postCodeZip' => $args['billing_postcode'],
                            'countryAlphaCode' => $args['billing_country'],
                        ]
                    ]
                ];
                
                if( $this->enable_tokenization == 'yes' ){
                    $configuration['tokenization']['merchantShopperReference'] = $order_username;
                    $configuration['tokenization']['tokenRegistrationProcess'] = 'Optional';
                }

                if ($truevo_gateway_settings['debug_log'] == 'yes') {
                    $log->add('truevo_gateway', 'Truevo payments request prepared and signed. ' . print_r($args, true));
                }

                $configuration = json_encode($configuration);
            }

            $url_user_on_success = add_query_arg(
                array(
                    'response_id' => $order->get_id(),
                    'response_secret' => md5("response_id" . $order->get_id())),
                site_url('/wc-api/truevo_gateway'
                )
            );

            ob_start();

            include_once 'templates/widget.php';

            echo ob_get_clean();
        }
        /**
         * truevo_gateway_webhook function.
         *
         * @access public
         * @return void
         */
        public function truevo_gateway_webhook()
        {

            if (!check_ajax_referer('truevo-security-nonce', 'security', false)) {
                wp_send_json_error('Invalid security token sent.');
                wp_die();
            }

            $order_id = sanitize_key($_GET['response_id']);
            $response_secret = sanitize_text_field($_GET['response_secret']);
            $order = wc_get_order($order_id);

            $truevo_order_flag = isset($_POST['truevo-order-flag']) ? $_POST['truevo-order-flag'] : '';
            $truevo_order_values = isset($_POST['truevo-order-values']) ? $_POST['truevo-order-values'] : '';

            if (md5("response_id" . $order_id) == $response_secret) {
                $truevo_order_values = stripslashes($truevo_order_values);
                $truevo_order_values = json_decode($truevo_order_values, true);

                foreach ($truevo_order_values as $key => $truevo_order_value) {
                    $order->update_meta_data('truevo_payment_' . $key, $truevo_order_value);
                }

                $redirect_url = $order->get_checkout_payment_url();

                if ($truevo_order_flag == "declined") {

                    $order->add_order_note('Truevo Payments: The payment was declined.', false);
                    wc_add_notice(__('Truevo Payments: Your payment failed. Please try again or use a different card.', 'truevo-gateway-woocommerce'), 'error');
                }

                if ($truevo_order_flag == "failed") {

                    $order->add_order_note('Truevo Payments: Payment Failed', false);
                    wc_add_notice(__('Truevo Payments: Your payment failed. Please try again or use a different card.', 'truevo-gateway-woocommerce'), 'error');
                }

                if ($truevo_order_flag == "completed") {

                    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');

                    if ($truevo_gateway_settings['enable_authorization'] == 'yes') {

                        $order->update_status('on-hold');

                        $order->add_order_note('The payment was pre-authorized. Please capture the payment when ready.', false);
                        $order->update_meta_data('truevo_payment_transaction_type', 'authorize');

                        // Add hidden meta data for getting amount to capture (starting with "_" will hide the meta data on the edit order page)
                        $order->update_meta_data('_truevo_payment_preauth_currency', $order->get_currency());
                        $order->update_meta_data('_truevo_payment_preauth_amount', $order->get_total());

                    } else {
                        $order->update_status('processing');
                        $order->reduce_order_stock();

                        $order->add_order_note('Truevo Payments: The payment was successfully completed.', false);
                        $order->update_meta_data('truevo_payment_transaction_type', 'sale');
                    }

                    // Empty cart
                    WC()->cart->empty_cart();

                    $redirect_url = $this->get_return_url($order);

                }

                $order->save();

                wp_redirect($redirect_url);
            }
        }

        /**
         * payment_fields function.
         *
         * @access public
         * @return void
         */
        public function payment_fields()
        {

            // ok, let's display some description before the payment form
            if ($this->description) {
                // you can instructions for test mode, I mean test card numbers etc.
                if ($this->testmode == 'test') {
                    $this->description .= ' TEST MODE ENABLED. In test mode, you can use the card numbers listed in <a href="https://truevo.com/payment-plugins/woocommerce-payment-plugin/" target="_blank" rel="noopener noreferrer">Documentation</a> or 5522000000000750.';
                    $this->description = trim($this->description);
                }
                // display the description with <p> tags etc.
                echo wpautop(wp_kses_post($this->description));
            }

        }

        /**
         * payment_scripts function.
         *
         * @access public
         * @return void
         */
        public function payment_scripts()
        {

            // we need JavaScript to process a token only on cart/checkout pages, right?
            if (!is_cart() && !is_checkout() && !isset($_GET['pay_for_order'])) {
                return;
            }

            // if our payment gateway is disabled, we do not have to enqueue JS too
            if ('no' === $this->enabled) {
                return;
            }

            // no reason to enqueue JavaScript if API keys are not set
            if (empty($this->merchant_id) || empty($this->terminal_id)) {
                return;
            }

            wp_enqueue_style("woocommerce-truevo-widget", plugins_url('assets/css/truevo-widget.css', __FILE__), [], strtotime('now'));
            wp_enqueue_script("woocommerce-truevo-apple-pay", "https://applepay.cdn-apple.com/jsapi/v1/apple-pay-sdk.js");

            if ($this->testmode === 'live') {
                wp_enqueue_script("woocommerce-truevo-merchant", plugins_url('assets/js/merchant.minified.js', __FILE__), [], strtotime('now'), true);
            } else {
                wp_enqueue_script("woocommerce-truevo-merchant", "https://sandbox-truevo-drop-in-hosted-payment-widget.s3.eu-west-1.amazonaws.com/merchant.minified.js", [], null, true);
            }
        }

        /**
         * process_refund function.
         *
         * @access public
         *
         * @param mixed $order_id
         * @param mixed $amount (default: null)
         * @param string $reason (default: '')
         *
         * @return void
         */
        public function process_refund($order_id, $amount = null, $reason = '')
        {

            $refund_mode = 'refund';

            if ($reason === 'Reverse_Full_Payment') {
                $refund_mode = 'reverse';
            }

            $order = wc_get_order($order_id);

            $truevo_payment_id = get_post_meta($order_id, 'truevo_payment_transactionId', true);

            if (!empty($truevo_payment_id)) {

                $order_reference = $order->get_id();

                $data_info = array(
                    'transactionUniqueIdentifier' => $truevo_payment_id,
                    'amount' => $amount,
                    'merchantReference' => "$order_reference",
                    'paymentReference' => $this->store_name,
                    'billing_address' => $order->get_billing_address_1(),
                    'billing_address_2' => $order->get_billing_address_2(),
                    'billing_city' => $order->get_billing_city(),
                    'billing_state' => $order->get_billing_state(),
                    'billing_country' => $order->get_billing_country(),
                    'billing_postcode' => $order->get_billing_postcode(),
                    'billing_first_name' => $order->billing_first_name,
                    'billing_last_name' => $order->billing_last_name,
                    'delivery_address' => $order->get_shipping_address_1(),
                    'delivery_address_2' => $order->get_shipping_address_2(),
                    'delivery_city' => $order->get_shipping_city(),
                    'delivery_country' => $order->get_shipping_country(),
                    'delivery_postcode' => $order->get_shipping_postcode(),
                    'delivery_state' => $order->get_shipping_state(),
                    'email' => $order->get_billing_email(),
                    'user_ip' => truevo_get_client_ip(),
                );
                if ($this->debug_log == 'yes') {
                    $this->log->add('truevo_gateway', 'Truevo payments refund/reverse request. ' . print_r($data_info, true));
                }

                $process_refund = $this->truevoapi->process_truevo_refund($data_info, $refund_mode);

                if ($this->debug_log == 'yes') {
                    $this->log->add('truevo_gateway', 'Truevo payments refund response. ' . print_r($process_refund, true));
                }
                $rejected = true;

                if ($process_refund['code'] == '200' && !empty($process_refund['result'])) {

                    if ($process_refund['result']->outcome->status === 'RECEIVED') {

                        $transaction_id = $process_refund['result']->transactionUniqueIdentifier;

                        $refund_final_response = $this->truevoapi->poll_transaction_status($transaction_id, $this->debug_log);

                        $refund_final_status = $refund_final_response['status'] ?? '';
                        $refund_final_message = $refund_final_response['reason'] ?? '';

                        if ($this->debug_log == 'yes') {
                            $this->log->add('truevo_gateway', 'Truevo payments refund status : ' . $refund_final_status);
                        }
                        if ($refund_final_status === 'COMPLETE') {

                            $order->add_order_note(sprintf('Truevo Payments : The payment %s was successfully refunded.', $amount), false);

                            update_post_meta($order_id, 'truevo_payment_refund_status', $refund_final_status);

                            $rejected = false;
                        } else {

                            update_post_meta($order_id, 'truevo_payment_refund_status', $refund_final_status);

                            if ($refund_final_message) {
                                $order->add_order_note($refund_final_message, false);
                            }

                        }


                    }
                }

                $order->save();

                if ($rejected) {

                    $order->add_order_note(sprintf('Truevo Payments: The refund was declined.', $amount), false);
                    return false;
                } else {
                    return true;
                }
            }

            $order->add_order_note("Truevo Payments: The refund could not be processed.", false);

            return false;
        }

        /**
         * process_payment function.
         *
         * @access public
         *
         * @param mixed $order_id
         *
         * @return void
         */
        public function process_payment($order_id)
        {

            $order = wc_get_order($order_id);

            return array(
                'result' => 'success',
                'redirect' => $order->get_checkout_payment_url(true)
            );
        }
    }
}

add_action('wp_ajax_nopriv_truevo_gateway_reverse_items', 'truevo_gateway_reverse_items');
add_action('wp_ajax_truevo_gateway_reverse_items', 'truevo_gateway_reverse_items');

/**
 * truevo_gateway_reverse_items function.
 *
 * @access public
 * @return void
 */
function truevo_gateway_reverse_items()
{

    if (!check_ajax_referer('truevo-admin-nonce', 'security', false)) {

        wp_send_json_error('Invalid security token sent.');
        wp_die();
    }


    if (isset($_POST['order_id']) && $_POST['order_id'] != "") {

        $log = new WC_Logger();

        $order_id = $_POST['order_id'];
        $order = wc_get_order($order_id);

        if (!is_a($order, 'WC_Order')) {
            return;
        }

        $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');


        if (!empty($truevo_gateway_settings)) {

            // Get Items
            $order_items = $order->get_items();

            // Refund Amount
            $refund_amount = 0;

            // Prepare line items which we are refunding
            $line_items = array();

            if ($order_items = $order->get_items()) {

                foreach ($order_items as $item_id => $item) {

                    $item_meta = $order->get_item_meta($item_id);

                    $product_data = wc_get_product($item_meta["_product_id"][0]);

                    $item_ids[] = $item_id;
                    $tax_data = $item_meta['_line_tax_data'];
                    $refund_tax = 0;

                    if (is_array($tax_data[0])) {

                        $refund_tax = array_map('wc_format_decimal', $tax_data[0]);

                    }

                    $refund_amount = wc_format_decimal($refund_amount) + wc_format_decimal($item_meta['_line_total'][0]);

                    $line_items[$item_id] = array('qty' => $item_meta['_qty'][0], 'refund_total' => wc_format_decimal($item_meta['_line_total'][0]), 'refund_tax' => $refund_tax);
                }
            }

            $refund = wc_create_refund(array(
                'amount' => $refund_amount,
                'reason' => "Reverse_Full_Payment",
                'order_id' => $order_id,
                'line_items' => $line_items,
                'refund_payment' => true
            ));

            if (is_wp_error($refund)) {
                $error_string = $refund->get_error_message();
                wp_send_json_error(json_encode($error_string));
            } else {
                if ($truevo_gateway_settings['debug_log'] == 'yes') {
                    $log->add('truevo_gateway', 'Truevo reverse processed. ' . print_r($refund, true));
                }
                wp_send_json_success();
            }
        }
    }

    wp_die();
}

add_action('admin_enqueue_scripts', 'truevo_backend_scripts');

/**
 * payment_scripts function.
 *
 * @access public
 * @return void
 */
function truevo_backend_scripts()
{
    wp_register_script('woocommerce-truevo-admin', plugins_url('assets/js/truevo-widget-admin.js', __FILE__), array('jquery'), '1.0.0');
    wp_localize_script(
        'woocommerce-truevo-admin',
        'truevo_ajax_object',
        [
            'security' => wp_create_nonce('truevo-admin-nonce'),
        ]
    );
    wp_enqueue_script('woocommerce-truevo-admin');
}

/**
 * truevo_get_client_ip function.
 *
 * @access public
 * @return void
 */
function truevo_get_client_ip()
{
    $ipAddress = isset($_SERVER['REMOTE_ADDR']) ? trim($_SERVER['REMOTE_ADDR']) : null;

    // Check IP address is valid and not localhost
    if (filter_var($ipAddress, FILTER_VALIDATE_IP) && (substr($ipAddress, 0, 4) != '127.' && $ipAddress != '::1')) {
        return $ipAddress;
    }

    return null;
}

/**
 * truevo_api_connection function
 *
 * @return false|Truevo_API_Wrapper
 */
function truevo_api_connection()
{
    $truevo_gateway_settings = get_option('woocommerce_truevo_gateway_settings');
    if (!empty($truevo_gateway_settings)) {
        $testmode = ('yes' === $truevo_gateway_settings['testmode']) ? 'test' : 'live';
        $merchant_id = ($testmode == 'test') ? $truevo_gateway_settings['test_merchant_id'] : $truevo_gateway_settings['live_merchant_id'];
        $terminal_id = ($testmode == 'test') ? $truevo_gateway_settings['test_terminal_id'] : $truevo_gateway_settings['live_terminal_id'];

        $api_url = ($testmode == 'test') ? 'https://sandbox.unify.truevo.com' : 'https://unify.truevo.com';

        $token_authorization = $truevo_gateway_settings['token_authorization'];


        $truevoapi = new Truevo_API_Wrapper($api_url, $token_authorization, $merchant_id, $terminal_id);

        return $truevoapi;
    }

    return false;
}